{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 17](https://adventofcode.com/2019/day/17) puzzle input is an IntCode program (mine is [here](puzzle_input/day17.txt)) which will output a camera's viewpoint of a twisting, turning \"scaffold\" with a little robot at one end. Part 1 involves finding the locations of all places where this scaffold intersects another. Part 2 involves \"programming\" the robot to make its way over the scaffold from start to finish (the commands are quite limited) and finding the output value at the end of this."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "..............##########^..............................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "......#########........................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "#######................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#...............................................#######\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#########.......................................#......\n",
      "........#.......................................#......\n",
      "........#.............#########.........#########......\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.#########...#.......#.........#..............\n",
      "........#.#.......#...#.......#.........#..............\n",
      "........#########.#...###########.......#..............\n",
      "..........#.....#.#...........#.#.......#..............\n",
      "..........#.....#.#...........###########..............\n",
      "..........#.....#.#.............#......................\n",
      "..........#.#######.............#......................\n",
      "..........#.#...#...............#......................\n",
      "..........#######...........#######....................\n",
      "............#...............#...#.#....................\n",
      "............#.............#######.#....................\n",
      "............#.............#.#.....#....................\n",
      "............#.#######.....#.#.....#....................\n",
      "............#.#.....#.....#.#.....#....................\n",
      "............###########...#.#.....#....................\n",
      "..............#.....#.#...#.#.....#....................\n",
      "..............#.....#.#...#########....................\n",
      "..............#.....#.#.....#..........................\n",
      "..............#.....#########..........................\n",
      "..............#.......#................................\n",
      "..............#########................................\n",
      "10064\n"
     ]
    }
   ],
   "source": [
    "puzzle_input = open(\"./puzzle_input/day17.txt\", \"r\").read()\n",
    "code = puzzle_input.split(\",\")\n",
    "cpu = IntCodeCpu(code)\n",
    "\n",
    "input_buffer = []\n",
    "output_buffer = []\n",
    "cpu.start(input_buffer, output_buffer)\n",
    "\n",
    "char_space = ord(\".\")\n",
    "char_scaffold = ord(\"#\")\n",
    "char_newline = ord(\"\\n\")\n",
    "char_robot_up = ord(\"^\")\n",
    "char_robot_down = ord(\"v\")\n",
    "char_robot_left = ord(\"<\")\n",
    "char_robot_right = ord(\">\")\n",
    "\n",
    "def is_scaffold(image, row, col):\n",
    "    return image[row][col] == char_scaffold\n",
    "    \n",
    "def parse_ascii_output(raw_image):\n",
    "    image_by_row = []\n",
    "    current_row = []\n",
    "    \n",
    "    for c in raw_image:\n",
    "        if c == char_newline:\n",
    "            if current_row:\n",
    "                image_by_row.append(current_row)\n",
    "                current_row = []\n",
    "        else:\n",
    "            current_row.append(c)\n",
    "\n",
    "    return image_by_row\n",
    "    \n",
    "    \n",
    "def print_ascii_output(image):\n",
    "    output = []\n",
    "    for row in image:\n",
    "        row_str = []\n",
    "        \n",
    "        for c in row:\n",
    "            row_str.append(chr(c))\n",
    "        \n",
    "        print( \"\".join(row_str))\n",
    "    \n",
    "image_by_row = parse_ascii_output(cpu.output_buffer)\n",
    "\n",
    "total = 0\n",
    "height = len(image_by_row)\n",
    "\n",
    "for i, row in enumerate(image_by_row):\n",
    "    width = len(row)\n",
    "    \n",
    "    for j, digit in enumerate(row):\n",
    "        if not is_scaffold(image_by_row, i, j):\n",
    "            continue\n",
    "        \n",
    "        scaffold_surroundings = [\n",
    "            i > 0 and is_scaffold(image_by_row, i-1, j),          # scaffold_above\n",
    "            j > 0 and is_scaffold(image_by_row, i, j-1),          # scaffold_left\n",
    "            1 + j < width and is_scaffold(image_by_row, i, j+1),  # scaffold_right\n",
    "            1 + i < height and is_scaffold(image_by_row, i+1, j), # scaffold_below\n",
    "        ]\n",
    "        \n",
    "        is_scaffold_join = len([ s for s in scaffold_surroundings if s ]) >= 3\n",
    "        \n",
    "        if is_scaffold_join:\n",
    "            total += (i * j)\n",
    "\n",
    "print_ascii_output(image_by_row)\n",
    "\n",
    "print(total)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Part of this involved determining the set of three functions that could be used to traverse the scaffold - I ended up doing this by hand:\n",
    "    \n",
    "![my notes for Day 17](img/day-17-notes.jpg \"my notes for Day 17\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "..............##########^..............................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "......#########........................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "#######................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#...............................................#######\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#########.......................................#......\n",
      "........#.......................................#......\n",
      "........#.............#########.........#########......\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.#########...#.......#.........#..............\n",
      "........#.#.......#...#.......#.........#..............\n",
      "........#########.#...###########.......#..............\n",
      "..........#.....#.#...........#.#.......#..............\n",
      "..........#.....#.#...........###########..............\n",
      "..........#.....#.#.............#......................\n",
      "..........#.#######.............#......................\n",
      "..........#.#...#...............#......................\n",
      "..........#######...........#######....................\n",
      "............#...............#...#.#....................\n",
      "............#.............#######.#....................\n",
      "............#.............#.#.....#....................\n",
      "............#.#######.....#.#.....#....................\n",
      "............#.#.....#.....#.#.....#....................\n",
      "............###########...#.#.....#....................\n",
      "..............#.....#.#...#.#.....#....................\n",
      "..............#.....#.#...#########....................\n",
      "..............#.....#.#.....#..........................\n",
      "..............#.....#########..........................\n",
      "..............#.......#................................\n",
      "..............#########................................\n",
      "Main:\n",
      "Function A:\n",
      "Function B:\n",
      "Function C:\n",
      "Continuous video feed?\n",
      "..............###########..............................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "..............#........................................\n",
      "......#########........................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "......#................................................\n",
      "#######................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#......................................................\n",
      "#...............................................######>\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#...............................................#......\n",
      "#########.......................................#......\n",
      "........#.......................................#......\n",
      "........#.............#########.........#########......\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.............#.......#.........#..............\n",
      "........#.#########...#.......#.........#..............\n",
      "........#.#.......#...#.......#.........#..............\n",
      "........#########.#...###########.......#..............\n",
      "..........#.....#.#...........#.#.......#..............\n",
      "..........#.....#.#...........###########..............\n",
      "..........#.....#.#.............#......................\n",
      "..........#.#######.............#......................\n",
      "..........#.#...#...............#......................\n",
      "..........#######...........#######....................\n",
      "............#...............#...#.#....................\n",
      "............#.............#######.#....................\n",
      "............#.............#.#.....#....................\n",
      "............#.#######.....#.#.....#....................\n",
      "............#.#.....#.....#.#.....#....................\n",
      "............###########...#.#.....#....................\n",
      "..............#.....#.#...#.#.....#....................\n",
      "..............#.....#.#...#########....................\n",
      "..............#.....#.#.....#..........................\n",
      "..............#.....#########..........................\n",
      "..............#.......#................................\n",
      "..............#########................................\n",
      "1197725\n"
     ]
    }
   ],
   "source": [
    "main_routine = \"A,A,B,C,B,C,B,C,B,A\"\n",
    "function_a = \"L,10,L,8,R,8,L,8,R,6\"\n",
    "function_b = \"R,6,R,8,R,8\"\n",
    "function_c = \"R,6,R,6,L,8,L,10\"\n",
    "no_video = \"n\"\n",
    "\n",
    "def assemble_instruction_group(instructions):\n",
    "    int_instructions = [ ord(c) for c in instructions ]\n",
    "    assert len(int_instructions) <= 20\n",
    "    return int_instructions + [ ord(\"\\n\") ]\n",
    "\n",
    "\n",
    "input_buffer = assemble_instruction_group(main_routine) + \\\n",
    "    assemble_instruction_group(function_a) + \\\n",
    "    assemble_instruction_group(function_b) + \\\n",
    "    assemble_instruction_group(function_c) + \\\n",
    "    assemble_instruction_group(no_video)\n",
    "\n",
    "code = puzzle_input.split(\",\")\n",
    "cpu = IntCodeCpu(code)\n",
    "cpu.memory[0] = 2\n",
    "output_buffer = []\n",
    "cpu.start(input_buffer[::-1], output_buffer)\n",
    "\n",
    "print_ascii_output(parse_ascii_output(cpu.output_buffer))\n",
    "\n",
    "print(output_buffer[-1])"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
